/*
 * utdstool.c
 *	writen by KIYOHARA Takashi <kiyohara@soum.co.jp>
 *	$SOUM$
 */


#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <linux/limits.h>

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define TDS_DEBUG

#include "../include/u3tdsio.h"

#define DEFAULT_UTDS_DEVICE "/dev/u3tds0"


struct key_value {
	char *key;  
	void *value;
	void *arg;
	int key_len;
};


char path[PATH_MAX] = DEFAULT_UTDS_DEVICE;
int fd;
int utdsctl_argc;
char **utdsctl_argv;


void usage();
static void init_key_value(struct key_value *);
static struct key_value *lookup_key_value(struct key_value *, const char *);
static struct key_value *lookup_value_key(struct key_value *, const void *);
static inline struct key_value *get_tagged_token(struct key_value *);
static int cmd_read(void *);
static int cmd_write(void *);
static int cmd_init(void *);
static int cmd_dump(void *);
static int cmd_debug(void *);
static int utdsctl();
static void regdump(unsigned char[]);


static const char separator[] = {' ', '\f', '\n', '\r', '\t', '\v', '\0'};

static struct key_value cmd_desc[] = {
	{ "read",	cmd_read,	(void *)NULL	},
	{ "write",	cmd_write,	(void *)NULL	},
	{ "init",	cmd_init,	(void *)NULL	},
	{ "dump",	cmd_dump,	(void *)NULL	},
	{ "debug",	cmd_debug,	(void *)NULL	},
	{ NULL,		NULL,		(void *)NULL	},
};


void
usage()
{
	printf("utdsctl: [-d node] command...\n");
	exit(EXIT_SUCCESS);
}

static void
init_key_value(struct key_value *table)
{   
	struct key_value *kvp;

	assert(table != NULL);

	/* fill key_len field of each key_value entry */
	for (kvp = table; kvp->key != NULL; kvp++)
		kvp->key_len = strlen(kvp->key) + 1;

	return;
}   

static struct key_value *
lookup_key_value(struct key_value *table, const char *key)
{
	struct key_value *kvp;

	assert(table != NULL);
	assert(key != NULL);

	for (kvp = table; kvp->key != NULL; kvp++)
		if (strncasecmp(key, kvp->key, kvp->key_len) == 0)
			return (kvp);
	return (NULL);
}

static struct key_value *
lookup_value_key(struct key_value *table, const void *value)
{
	struct key_value *kvp;

	assert(table != NULL);

	for (kvp = table; kvp->key != NULL; kvp++)
		if(value == kvp->value)
			return (kvp);
	return (NULL);
}

static inline char *
token_to_value(char *token, long *valp)
{
	double fval;
	unsigned long ulval;
	char factor[8], unit[8];

	factor[0] = unit[0] = '\0';

	if (sscanf(token, "%lf%1[KkMm]%1[bB]", &fval, factor, unit) >= 2) {
		double kbytes = (unit[0] == '\0' ? 1000.0 : 1024.0);

		switch (factor[0]) {
		case 'K': case 'k':
			*valp = (long)(fval * kbytes);
			break;
		case 'M': case 'm':
			*valp = (long)(fval * kbytes * kbytes);
			break;
		default:
			return (NULL);
        	}
	} else if (sscanf(token, "%lli", &ulval) == 1)
		*valp = ulval;
	else
		return (NULL);
	return (token);
}

static char *
get_numeric_token(long *valp)
{
	char *token;

	if (utdsctl_argc < 1 ||
	    (utdsctl_argc > 0 && (token = *utdsctl_argv) == NULL))
		return (NULL);
	utdsctl_argc --;
	utdsctl_argv ++;

	return (token_to_value(token, valp));
}

static inline struct key_value *
get_tagged_token(struct key_value *table)
{   
	struct key_value *kvp;
	char *token;

	if (utdsctl_argc < 1 ||
	    (utdsctl_argc > 0 && (token = *utdsctl_argv) == NULL))
		return (NULL);
	utdsctl_argc --;
	utdsctl_argv ++;
	if ((kvp = lookup_key_value(table, token)) == NULL)  
		return (NULL);
	return (kvp);
}


static int
cmd_read(void *arg)
{
	struct tdsio_regio regio;
	long address;

	if (utdsctl_argc != 1) {
		fprintf(stderr, "Invalid argument\n");
		goto usage;
	}
	if (get_numeric_token(&address) == NULL) {
		fprintf(stderr, "address value expected\n");
		goto usage;
	}

	regio.address = address;

	if (ioctl(fd, TDSIO_REGIO_READ, &regio) < 0)
		err(EXIT_FAILURE, "%s: TDSIO_REGIO_READ", path);
	printf("%s: cmd_read address=0x%02x, value=0x%02x\n",
	     path, regio.address, (uint8_t)regio.value);

	return (0);

usage:
	fprintf(stderr, "usage: read <address>\n");
	fprintf(stderr, "    <address>: registor address\n");
	return (EINVAL);
}

static int
cmd_write(void *arg)
{
	struct tdsio_regio regio;
	long address, value;

	if (utdsctl_argc != 2) {
		fprintf(stderr, "Invalid argument\n");
		goto usage;
	}
	if (get_numeric_token(&address) == NULL) {
		fprintf(stderr, "address value expected\n");
		goto usage;
	}
	if (get_numeric_token(&value) == NULL) {
		fprintf(stderr, "value expected\n");
		goto usage;
	}

	regio.address = address;
	regio.value = value;

	if (ioctl(fd, TDSIO_REGIO_WRITE, &regio) < 0)
		err(EXIT_FAILURE, "%s: TDSIO_REGIO_WRITE", path);
	printf("%s: cmd_write address=0x%02x, value=0x%02x\n",
	     path, regio.address, regio.value);

	if (ioctl(fd, TDSIO_REGIO_READ, &regio) < 0)
		err(EXIT_FAILURE, "%s: TDSIO_REGIO_READ", path);
	printf("%s: cmd_read address=0x%02x, value=0x%02x\n",
	     path, regio.address, regio.value);

	return (0);

usage:
	fprintf(stderr, "usage: write <address> <value>\n");
	fprintf(stderr, "    <address>: registor address\n");
	fprintf(stderr, "    <value>: value\n");
	return (EINVAL);
}


static int
cmd_init(void *arg)
{

	if (utdsctl_argc != 0) {
		fprintf(stderr, "Invalid argument\n");
		goto usage;
	}

	if (ioctl(fd, TDSIO_INIT) < 0)
		err(EXIT_FAILURE, "%s: TDSIO_INIT", path);
	printf("%s: cmd_init successed\n", path);

	return (0);

usage:
	fprintf(stderr, "usage: init\n");
	return (EINVAL);
}


static int
cmd_dump(void *arg)
{
	struct tdsio_regio regio;
	long address, value;
	int i;
	unsigned char reg[0x20];

	if (utdsctl_argc != 0) {
		fprintf(stderr, "Invalid argument\n");
		goto usage;
	}

	for (i = 0; i < sizeof (reg); i++) {
		regio.address = i;
		if (ioctl(fd, TDSIO_REGIO_READ, &regio) < 0)
			err(EXIT_FAILURE, "%s: TDSIO_REGIO_READ", path);
		reg[i] = regio.value;
	}

	regdump(reg);

	return (0);

usage:
	fprintf(stderr, "usage: dump\n");
	return (EINVAL);
}


static int
cmd_debug(void *arg)
{
	uint32_t val;
	int request = 0;
	char *fmt;

	if (utdsctl_argc == 0) {
		request = TDSIO_GET_DEBUG;
		fmt = "%s: set debug %d\n";
	} else if (utdsctl_argc == 1) {
		request = TDSIO_SET_DEBUG;
		if (get_numeric_token((long *)&val) == NULL) {
			fprintf(stderr, "value expected\n");
			goto usage;
		}
		fmt = "%s: current debug %d\n";
	} else {
		fprintf(stderr, "Invalid argument\n");
		goto usage;
	}

	if (ioctl(fd, request, &val) < 0)
		err(EXIT_FAILURE, "%s: cmd_debug", path);

	printf(fmt, path, val);
	return (0);

usage:
	fprintf(stderr, "usage: debug [val]\n");
	fprintf(stderr, "    val: set debug variable\n");
	return (EINVAL);
}


int
main(int argc, char **argv)
{
	char c;

	while((c = getopt(argc, argv, "d:")) != -1) {
		switch(c) {
		case 'd':
			strncpy(path, optarg, sizeof (path));
			break;

		default:
			usage();
			/* NOTREACHED */
		}
	}
	utdsctl_argc = argc - optind;
	utdsctl_argv = argv + optind;

	if ((fd = open(path, O_RDWR)) <= 0)
		err(EXIT_FAILURE, "%s", path);

	return (utdsctl(argc, argv));
}

static int
utdsctl()
{
	struct key_value *kvp;
	int (*func)(void *);
	int status;
	char *token;

	init_key_value(cmd_desc);

	if (utdsctl_argc < 1) {
		fprintf(stderr, "Invalid argument\n");
		return (EINVAL);
	}
	token = *utdsctl_argv;
	utdsctl_argc --;
	utdsctl_argv ++;
	if((kvp = lookup_key_value(cmd_desc, token)) != NULL) {
		func = kvp->value;
		status = func(kvp->arg);
	} else
		fprintf(stderr, "Invalid argument\n");
	return (status);
}


static void
regdump(unsigned char reg[])
{
	uint32_t v32;
	int i, j, k;
	char **str, *op;
	char *status[] = {
	    "",         "",         " 5MHz",    " PLLlck",
	    " RefIn",   " 1PPSIn",  " DataExst"," OnSmpl"   };
	char *filter[] = {
/* 2016.02.19 H.Ohkubo for 9277 */
//	    "Thru", "16M", "8M",  "4M",  "2M",  "?",   "?",    "?"  };
	    "Thru", "16M", "8M",  "4M",  "2M",  "32M",   "?",    "?"  };
	char *resolutionbit[] = { "1bit", "2bit", "4bit", "8bit" };
	char *fsample[] = {
	    "0.04M","0.1M", "0.2M", "0.5M", "1M",   "2M",   "4M",   "8M",
/* 2016.02.19 H.Ohkubo for 9277 */
//	    "16M",  "32M",  "64M",  "?",    "?",    "?",    "?",    "?"    };
	    "16M",  "32M",  "64M",  "128M",    "?",    "?",    "?",    "?"    };

	for (i = 0; i < 0x20; i++) {
		printf("0x%02x: 0x%02x	", i, reg[i]);

		switch (i) {
		case 0x00:
			if (i == 0)
				str = status;

			for (j = 0, k = 0; j < 8; j++)
				if ((reg[i] & (1 << (8 - j - 1))) != 0) {
					if (k != 0)
						printf(", ");
					printf("%s", str[j]);
					k = 1;
				}
			break;
		case 0x01:
		case 0x05:
			if (i == 0x01)
				op = "Write";
			else
				op = "Read";
			v32 = *((uint32_t *)(&reg[i]));
			v32 &= 0x01ffffff;
			printf(" Data Pointer (%s) %d (0x%x)", op, v32, v32);
			break;
		case 0x10:
			printf(" Filter %s, Resolution Bit %s, CH Mode %s",
			    filter[(reg[i] & 0x31) >> 3],
			    resolutionbit[(reg[i] & 0x06) >> 1],
			    reg[i] & 0x01 ? "4CH" : "1CH");
			break;
		case 0x11:
			printf(" Fsample %s", fsample[reg[i] & 0x0f]);
			break;
		case 0x14:
			v32 = *((uint32_t *)(&reg[i]));
			printf(" total seconds %d, total days %d",
			    v32 & 0x0001ffff, (v32 & 0x07fc0000) >> 18);
			break;
		case 0x18:
			printf(" year 20%02d", reg[i]);
			break;
		case 0x1c:
		case 0x1d:
		case 0x1e:
		case 0x1f:
			printf(" DC Offset %dCH %d", i - 0x1c + 1, reg[i]);
			break;
		default:
			break;
		}

		printf("\n");
	}
}
